"""
This module provides a set of classes to be subclassed by DAG generator
plug-ins.  A DAG generator plug-in is written as a Python module that
imports and subclasses each and every class in this module.  The subclasses
must each have the same name as in this module, for example the plug-in
must provide a class named DAGGenerator that is a subclass of the
DAGGenerator class in this module.  There is also a module-level function,
a replacement for which must be provided by the plug-in.  The resulting
Python module is then installed in the user's Python path, and the name of
the module is given to the daemon in its configuration file.
"""

# $Id: DAGGenerator.py,v 1.13 2005/09/12 19:51:24 kipp Exp $

#
# Preamble.
#

from onasys import onasysd


#
# Base class for DAG generator interfaces.
#

class DAGGeneratorError(onasysd.OnasysError):
	"""
	Base class for DAG generator exceptions.
	"""
	pass

class DAGGeneratorConfig(onasysd.Config):
	"""
	Base class for DAG generator configuration file interfaces.
	Generally a subclass will override the defaults() method, and the
	parse() method, and optionally the validate() method.
	"""
	section = "daggenerator"

class DAGGeneratorParameters(object):
	"""
	Base class for DAG generator parameter objects.  To the daemon,
	these are opaque objects which it will never look inside.  The role
	of this class is to provide a place-holder object used to represent
	an arbitrary set of parameters to be passed into the DAG generator.

	A subclass will typically provide an __init__() override for its
	own use in creating one of these things, and also a __repr__()
	override.  The daemon will call the __repr__() method to print
	messages in the log file.
	"""
	pass

def DAGGeneratorParameterList(clock, config, available, used):
	"""
	Module-level function for constructing a list of DAG generator
	parameters.  User-supplied DAG generator modules are expected to
	provide an override for this function that will return a list of
	DAGGeneratorParameters objects.  For each object in this list, the
	daemon will create a DAGGenerator object by invoking the
	DAGGenerator class' __init__() method with the object passed as its
	params argument.

	When called, config will be a DAGGeneratorConfig object (i.e. the
	contents of the "daggenerator" section of the daemon's config
	file); available will be a dictionary of segmentlist objects as
	produced by the SegGenerator plug-in;  and used will be a
	dictionary of segmentlist objects as produced by calls to the
	generate() method of the DAGGenerator class.

	Note that the daemon does not impose any restrictions on the
	dictionary produced by the SegGenerator plug-in.  It is left as an
	exercise to the plug-in author to ensure that this function is
	compatible with the dictionary of segmentlists produced by the
	SegGenerator.
	"""
	return []

class DAGGenerator(object):
	"""
	Class representing the construction of a DAG.  One of these objects
	will be created for each DAG to be constructed.  Generally, a
	subclass will provide an override for the generate() method, and
	possibly an override for the dag_name() method.
	"""
	def __init__(self, tag, config, params):
		"""
		Construct a DAGGenerator object.  tag is a text string used
		to identify the analysis being performed, config is a
		DAGGeneratorConfig object (i.e. the contents of the
		"daggenerator" section of the daemon's config file), and
		params is a DAGGeneratorParameters object (i.e. one element
		of the list produced by the DAGGeneratorParameterList()
		function).
		"""
		self.tag = tag
		self.config = config
		self.params = params

	def dag_name(self):
		return self.tag + ".dag"

	def generate(self):
		"""
		Generate the DAG represented by this object.  When this
		method is called, the current working directory will be set
		to an empty directory, into which it is expected a DAG will
		be placed.  The .dag file itself must have the name
		returned by the dag_name() method.

		When complete, this method must return a dictionary of
		segmentlist objects describing the segments that are "used"
		by this DAG.  The daemon maintains a dictionary of used
		segmentlists, each element of which is the union of the
		corresponding elements from all previous calls to
		generate() methods.  There are no pre-defined elements in
		this dictionary, rather its contents are defined entirely
		by the output of this function.  The resulting dictionary
		is passed as the used parameter to the
		DAGGeneratorParameterList() function.
		"""
		newlyused = {}
		return newlyused
