"""
This module provides a set of classes to be subclassed by segment generator
plug-ins.  A segment generator plug-in is written as a Python module that
imports and subclasses each and every class in this module.  The subclasses
must each have the same name as in this module, for example the plug-in
must provide a class named SegGenerator that is a subclass of the
SegGenerator class in this module.  The resulting Python module is then
installed in the user's Python path, and the name of the module is given to
the daemon in its configuration file.
"""

# $Id: SegGenerator.py,v 1.3 2005/09/12 19:48:10 kipp Exp $

#
# Preamble.
#

from glue import segments
from onasys import onasysd


#
# Base classes for segment generator interfaces.
#

class SegGeneratorError(onasysd.OnasysError):
	"""
	Base class for segment generator exceptions.
	"""
	pass

class SegGeneratorConfig(onasysd.Config):
	"""
	Base class for segment generator configuration file interfaces.
	Generally a subclass will override the defaults() method, and the
	parse() method, and optionally the validate() method.
	"""
	section = "seggenerator"

class SegGenerator(object):
	"""
	Class defining a segment generator.  The daemon will create one
	instance of this class which will then be called upon to update
	the list(s) of segments to be passed to the DAG generator plug-in.
	"""

	def __init__(self, config):
		"""
		Construct a SegGenerator object.  The config parameter is a
		SegGeneratorConfig object (i.e. the contents of the
		[seggenerator] section of the configuration file).
		"""
		self.config = config

	def get_state(self):
		"""
		Returns an object describing the internal state of the
		segment generator.  This object will be preserved in the
		daemon's state file and will be passed to the set_state()
		method if recovered from the state file on restart.
		"""
		pass

	def set_state(self, state):
		"""
		Restores the segment generator's internal state using the
		contents of state, which is an object previously returned
		from the get_state() method and saved in the daemon's state
		file.
		"""
		pass

	def generate(self, clock):
		"""
		Returns a dictionary of segmentlist objects, with one
		element for each type of segment available for analysis.
		The individual segmentlists need not be complete.  The
		daemon maintains a dictionary whose elements are each the
		union of the corresponding elements from all previous calls
		to this method, so this method need only output the
		segments that are new.

		clock is the daemon's onasysd.Clock instance.

		The daemon's dictionary (containing the union of all past
		output from this method) is passed as the "available"
		parameter when invoking the DAG generator's
		DAGGeneratorParameterList() function, and so the names of
		the elements in the dictionrry produced by this method must
		be compatible with the names expected by the DAG generator.
		It is left as an exercise to the authors of both plug-ins
		to sort this out amongst themselves.
		"""
		return {}
